import java.io.*;
import java.net.*;

public class ChatClient extends ChatClientBase
{
    private 
    int                 m_nMyId             = 0;
    
    private
    String              m_sNick             = null;
    
    private
    double              m_fVersion          = 2.1;
        
    private
    ChatClientInfoList  m_cClients          = new ChatClientInfoList();

    private
    Chat                m_cChat             = null;
    
    public ChatClient( String sHost, int nPort, 
                       String sNick, Chat cChat )
    {
	    super( sHost, nPort );

		m_cChat = cChat;
        m_sNick = sNick;
        start();
    }
    
    void Destroy()
    {
        m_cChat     = null;
        CloseUp();
    }
    
    public void run()
    {
        if ( ! m_bStarted )
        {
            m_bStarted = true;
            Start();
        }
        super.run();
    }
  
    protected void Start()
    {
        try 
        {
            m_bRunning = true;
            
            super.Start();
            
            synchronized ( m_cInput ) 
            {
	            synchronized ( m_cOutput ) 
                {
	    	        SendMyVersion();
                
	    	        if ( m_cInput.readShort() == EChatCommand._SERVER_VERSION )
	    	            RecvServerVersion();
	    	        else
	    	        {
                        InputReadAll();
	    	        }
	            }
	        }
	    
	        SendMyInfo();
            SendRequestStoredMsgs();
	    }
	    catch (UnknownHostException e1) 
        {
            OutputMsg( "Can't connect. Unknown host "+m_sHost, EChatMessageType._ERROR );
            m_bRunning = false;
	        return;
	    }
	    catch (IOException e2) 
        {
            OutputMsg( "Can't connect. Chat server not found on "+m_sHost, EChatMessageType._ERROR );
            m_bRunning = false;
	        return;
	    }
	    catch (Exception e3)
	    {
            OutputMsg( "Can't connect.", EChatMessageType._ERROR );
            m_bRunning = false;
		    return;
	    }
    }
    
    
    protected void Disconnect(String sReason)
    {
        OutputMsg("Disconnected.",EChatMessageType._SYSTEM);
        try 
        {
	        m_bRunning = false;
            if ( m_cOutput != null )
	            SendRequestDisconnect( sReason );
	        CloseUp();
	    }
	    catch (IOException e) {}
    }
    
    protected void RecvFromServer() throws IOException
    {
	    short nCommand = EChatCommand._NONE;
	    synchronized ( m_cInput ) 
        {
	        m_cInput.mark(1024);

	        nCommand = m_cInput.readShort();
            //GameObject.CrashLog("Command "+nCommand+" m_bShutDown "+m_bShutDown+"\n");

	        if ( ! m_bRunning )
		        return;
	        
		    switch ( nCommand ) 
            {
                case EChatCommand._CHAT_MSG:
			        RecvChatMsg();
			        break;

                case EChatCommand._SPECIAL_MSG:
			        RecvSpecialMsg();
			        break;
			
		        case EChatCommand._STORED_MSGS:
			        RecvStoredMsgs();
			        break;
                        
                case EChatCommand._MSG_FROM_SERVER:
			        RecvMsgFromServer();
			        break;                        
                        
		        case EChatCommand._NEW_USER_CONNECTED:
			        RecvNewUserConnected();
			        break;
			
                case EChatCommand._USER_DISCONNECTED:
			        RecvUserDisconnected();
			        break;
                        
		        case EChatCommand._USER_INFO:
			        RecvUserInfo();
			        break;
					            
			
		        case EChatCommand._NOOP:
                case EChatCommand._PING:
			        break;
			
                case EChatCommand._USER_ACTIVITY:
			        RecvAndIgnoreUserActivity();
			        break;
                        
		        case EChatCommand._CHANNEL_LIST:
			        RecvAndIgnoreChannelList();
			        break;
			
		        case EChatCommand._INVITE_CHANNEL:
			        RecvAndIgnoreInviteChannel();
			        break;
			
		        case EChatCommand._USER_ENTER_CHANNEL:
			        RecvAndIgnoreUserEnterChannel();
			        break;		   
			
		        case EChatCommand._KICK_USER:
			        RecvAndIgnoreKickUser();
			        break;		   
			
		        case EChatCommand._BAN_USER:
			        RecvAndIgnoreBanUser();
			        break;		   
			
		        case EChatCommand._UNBAN_USER:
			        RecvAndIgnoreUnbanUser();
			        break;		   
			
		        case EChatCommand._DRAW_LINE:
			        RecvAndIgnoreLine();
			        break;
			
		        case EChatCommand._DRAW_RECT:
			        RecvAndIgnoreDrawRect();
			        break;
			
		        case EChatCommand._DRAW_OVAL:
			        RecvAndIgnoreDrawOval();
			        break;
			
		        case EChatCommand._DRAW_TEXT:
			        RecvAndIgnoreDrawText();
			        break;
			
		        case EChatCommand._DRAW_PICTURE:
			        RecvAndIgnoreDrawPicture();
			        break;
			
		        case EChatCommand._CLEAR_BOARD:
			        RecvAndIgnoreClearBoard();
			        break;
			
		        case EChatCommand._PAGE_USER:
			        RecvAndIgnorePageUser();
			        break;
			
		        case EChatCommand._ERROR:
			        RecvAndIgnoreError();
			        break;
			
		        default:
		            InputReadAll();
		    }
	    }
    }

    
    
    ///////////////////////////////////////////////////
    ///////////////////////////////////////////////////
    ///////////// Send command to server //////////////
    ///////////////////////////////////////////////////
    ///////////////////////////////////////////////////
        
    protected void SendMyVersion() throws IOException
    {
	    synchronized ( m_cOutput )
	    {
		    m_cOutput.writeShort( EChatCommand._MY_VERSION );
		    m_cOutput.writeDouble( m_fVersion );
	    }
    }

    protected void SendMyInfo() throws IOException
    {
	    synchronized ( m_cOutput )
	    {
		    m_cOutput.writeShort( EChatCommand._MY_INFO );
		    m_cOutput.writeInt( 0 );
		    m_cOutput.writeUTF( m_sNick );
		    m_cOutput.writeUTF( "" );
		    m_cOutput.writeBoolean( false );
		    m_cOutput.writeUTF( "" );
            m_cOutput.writeUTF("en");
	    }
    }
    protected void SendChatMsg(String sText) throws IOException
    {
	    synchronized ( m_cOutput )
	    {
		    m_cOutput.writeShort( EChatCommand._CHAT_MSG );
		    m_cOutput.writeInt( m_nMyId );
	        m_cOutput.writeBoolean( false );
		    m_cOutput.writeShort( 0 );
		    m_cOutput.writeUTF( sText );
            m_cOutput.writeInt( 0 );
	    }
    }
    
    protected void SendRequestDisconnect(String sReason) throws IOException
    {
	    synchronized ( m_cOutput )
	    {
		    m_cOutput.writeShort( EChatCommand._REQUEST_DISCONNECT );
		    m_cOutput.writeInt( m_nMyId );
		    m_cOutput.writeUTF( sReason );
	    }
    }
        
    public void SendRequestStoredMsgs() throws IOException
    {
	    synchronized ( m_cOutput )
	    {
		    m_cOutput.writeShort( EChatCommand._REQUEST_STORED_MSGS );
		    m_cOutput.writeInt( m_nMyId );
	    }
    }
    
    ///////////////////////////////////////////////////
    ////////// END - Send command to server ///////////
    ///////////////////////////////////////////////////
    
    
    ///////////////////////////////////////////////////
    ///////////////////////////////////////////////////
    ////////// Receive command from server ////////////
    ///////////////////////////////////////////////////
    ///////////////////////////////////////////////////
    
    private void RecvServerVersion() throws IOException
    {
	    m_fVersion = m_cInput.readDouble();
    }
    
    private void RecvUserInfo() throws IOException
    {
	    int		nId		= 0;
	    String	sNick	= "";
	    
	    nId		= m_cInput.readInt();
	    sNick	= m_cInput.readUTF();

		m_cInput.readUTF();
	    m_cInput.readBoolean();
	    m_cInput.readUTF();
        m_cInput.readUTF();
	    
	    if ( sNick.equals( m_sNick ) )
        {
			m_nMyId = nId;
            SetConnectionState( EChatConnectionState._CONNECTED );
            OutputMsg("Connected as "+m_sNick,EChatMessageType._SYSTEM);
        }

        synchronized ( m_cClients )
        {
	        m_cClients.Add( new ChatClientInfo( nId, sNick ) );
        }
            
        if ( m_cChat != null )
             m_cChat.m_bClientsChanged = true;
    }

    private void RecvNewUserConnected() throws IOException
    {
	    String sNick = m_cInput.readUTF();
        OutputMsg("'"+sNick+"' connected",EChatMessageType._SYSTEM);
    }
    
    private void RecvUserDisconnected() throws IOException
    {
	    int     nId     = 0;
	    String  sReason = "";

	    nId     = m_cInput.readInt();
	    sReason = m_cInput.readUTF();
	    m_cInput.readFully(new byte[ m_cInput.available() ]);
        if ( sReason.equals(""))
		     sReason = "no reason";
        
	    if (( nId == m_nMyId) || (nId == 0))
	    {
            if ( sReason.startsWith("There is already a user called") )
            {
                Chat cChat = m_cChat;
                CloseUp();
                if ( cChat != null )
                    cChat.ScheduleReconnect();
            }
            else
            {
                OutputMsg("You have been disconnected",EChatMessageType._SYSTEM);
		        CloseUp();
            }
	    }
	    else
	    {
		    // Some other user disconnected.  Output a message that this
		    // user has left the chat.

		    ChatClientInfo cClientInfo = FindClientInfoById( nId );

		    if (cClientInfo != null)
            {
                String sNick = cClientInfo.GetNick();
                OutputMsg("'"+sNick+"' disconnected",EChatMessageType._SYSTEM);
                                
                synchronized ( m_cClients )
                {
	                m_cClients.Remove( cClientInfo );
                }
                
                if ( m_cChat != null )
                    m_cChat.m_bClientsChanged = true;
            }
        }
	}

    private void RecvChatMsg() throws IOException
    {
	    ChatClientInfo cSenderInfo  = null;
	    boolean        bPrivate     = false;
	    String         sMsg         = "";
	    

	    cSenderInfo = InputReadClient();
	    bPrivate    = m_cInput.readBoolean();
	    m_cInput.readShort();
	    sMsg        = m_cInput.readUTF();
        InputReadAndSkipRecipients();
        
	    if ( cSenderInfo != null )
	    {
            sMsg = cSenderInfo.GetNick() + "> " + sMsg;
	    }
    
        int nType = bPrivate ? 
                    EChatMessageType._PRIVATE_CHAT: EChatMessageType._CHAT;
        OutputMsg( sMsg, nType );
    }
    private void RecvSpecialMsg() throws IOException
    {
        ChatClientInfo cSenderInfo  = null;
	    String         sMsg         = "";
			
	    cSenderInfo = InputReadClient();
	    m_cInput.readInt();
	    sMsg        = m_cInput.readUTF();

	    if ( cSenderInfo != null )
        {
            sMsg = cSenderInfo.GetNick() + "> " + sMsg;
        }
        
        OutputMsg( sMsg, EChatMessageType._SPECIAL );
    }

    private void RecvMsgFromServer() throws IOException
    {
	    OutputMsg( m_cInput.readUTF(), EChatMessageType._FROM_SERVER );
    }
    
    private void RecvStoredMsgs() throws IOException
    {
	    short   nMsgCnt = 0;
	    nMsgCnt = m_cInput.readShort();

	    if ( nMsgCnt > 0 )
	        for ( int i = 0; i < nMsgCnt; ++ i )
	        {
                String  sNick   = "";
	            String  sMsg    = "";
		        sNick = m_cInput.readUTF();
		        sMsg  = m_cInput.readUTF();
                
                OutputMsg( sNick+"> "+sMsg, EChatMessageType._STORED );
            }
    }
    
        /////////////////////////////////////////
        /// Commands to receive and ignore    ///
        /////////////////////////////////////////
    
    private void RecvAndIgnoreChannelList() throws IOException
    {
	    int nRoomsCnt = 0;
	    
        nRoomsCnt = m_cInput.readShort();
	    
	    for ( int i = 0; i < nRoomsCnt; ++i )
	    {		    
		    m_cInput.readUTF();
		    m_cInput.readUTF();
		    m_cInput.readBoolean();
		    m_cInput.readBoolean();
		    
            int nUsers = m_cInput.readInt();
		    for ( int j = 0; j < nUsers; ++j )
		        m_cInput.readUTF();
	    }
    }

    private void RecvAndIgnoreInviteChannel() throws IOException
    {
	    InputReadClient();
        m_cInput.readUTF();
	    m_cInput.readInt();
    }
    private void RecvAndIgnoreUserEnterChannel() throws IOException
    {
	    m_cInput.readInt();
	    m_cInput.readUTF();
	    m_cInput.readBoolean();
	    m_cInput.readUTF(); 
	    m_cInput.readBoolean(); 
    }
    
    private void RecvAndIgnoreUserActivity() throws IOException
    {
	    InputReadClient();
	    m_cInput.readShort();
	    int nUsers = 0;
        nUsers = m_cInput.readInt();
	    for ( int i = 0; i < nUsers; ++i )
	        m_cInput.readInt();
    }
    
	private void RecvAndIgnoreKickUser() throws IOException
    {	        
		InputReadClient();
		m_cInput.readUTF();
		m_cInput.readInt();
    }
    
	private void RecvAndIgnoreBanUser() throws IOException
    {
		InputReadClient();
		m_cInput.readUTF();
		m_cInput.readInt();
    }
    
	private void RecvAndIgnoreUnbanUser() throws IOException
    {	        
		InputReadClient();
		m_cInput.readUTF();
		m_cInput.readInt();
    }
	
    			    
	private void RecvAndIgnoreLine() throws IOException
    {	     
		InputReadClient();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readShort();
		InputReadAndSkipRecipients();
    }
    		    
	private void RecvAndIgnoreDrawRect() throws IOException
    {	        
		InputReadClient();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readBoolean();
		InputReadAndSkipRecipients();
    }
    			    
	private void RecvAndIgnoreDrawOval() throws IOException
    {	        
		InputReadClient();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readBoolean();
		InputReadAndSkipRecipients();
    }
    			    
	private void RecvAndIgnoreDrawText() throws IOException
    {	        
		InputReadClient();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readShort();
		m_cInput.readUTF();
		InputReadAndSkipRecipients();
    }
    			    
	private void RecvAndIgnoreDrawPicture() throws IOException
    {
		int		nLen	= 0;
		byte	aData[] = null;
		InputReadClient();
		m_cInput.readShort();
		m_cInput.readShort();
		nLen = m_cInput.readInt();
		aData = new byte[ nLen ];
		int nRead = 0;
		while ( nRead < nLen )
			nRead += m_cInput.read( aData, nRead, nLen - nRead );
		InputReadAndSkipRecipients();
    }
    			    
	private void RecvAndIgnoreClearBoard() throws IOException
    {	        
		InputReadClient();
		InputReadAndSkipRecipients();
    }
    			    
	private void RecvAndIgnorePageUser() throws IOException
    {
		InputReadClient();
		InputReadAndSkipRecipients();
    }
    
    
    private void RecvAndIgnoreError() throws IOException
    {
	    ChatClientInfo cSenderInfo  = null;
	    short          nError       = 0;
	    
	    InputReadClient();
	    m_cInput.readShort();
        InputReadAndSkipRecipients();
    }
    ///////////////////////////////////////////////////
    /////// END - Receive command from server /////////
    ///////////////////////////////////////////////////
    
    
    ///////////////////////////////////////////////////
    ///////////////////////////////////////////////////
    //////////////// Read input stream ////////////////
    ///////////////////////////////////////////////////
    ///////////////////////////////////////////////////
    
    private ChatClientInfo InputReadClient() throws IOException
    {
	    int nId = 0;
	    nId = m_cInput.readInt();

	    return ( FindClientInfoById( nId ) );
    }
    
    private void InputReadAndSkipRecipients() throws IOException
    {
	    int nRecipientsCnt = m_cInput.readInt();
	    for ( int i = 0; i < nRecipientsCnt; ++ i )
	        m_cInput.readInt();
    }
    
    private byte[] InputReadAll() throws IOException
    {
        byte[] aAll = new byte[ m_cInput.available() ];
        m_cInput.readFully( aAll );
        return aAll;
    }
    
    private ChatClientInfo FindClientInfoById(int nId)
    {
        synchronized (m_cClients)
        {
            return m_cClients.FindClientInfoById( nId );
        }
    }
    
    ///////////////////////////////////////////////////
    ///////////// END - Read input stream /////////////
    ///////////////////////////////////////////////////
    
    
    
    protected void ConnectionLost()
    {
	    if ( m_bRunning )
	    {
            OutputMsg("Connection to server lost",EChatMessageType._SYSTEM);
		    CloseUp();
	    }
    }

    protected synchronized void CloseUp()
    {
        SetConnectionState( EChatConnectionState._DISCONNECTED );
        
	    m_bRunning = false;
        if ( m_cInput != null )
        {
	        try 
            {
	            m_cInput.close();
	        }
	        catch (IOException e1) {}
        }
	
	    try 
        {
            if ( m_cOutput != null )
            {
	            synchronized ( m_cOutput )
		        {
		            SendToServer();
		            m_cOutput.close();
		        }
            }
            if ( m_cSocket != null )
	            m_cSocket.close();
	    }
	    catch (IOException e2) {}
        
        m_cInput = null;
        m_cOutput = null;
        m_cSocket = null;
	    m_cClients.Clear();
        m_cChat = null;
	    System.gc();
    }
    
    private void SetConnectionState( int nState )
    {
        if ( m_cChat != null )
            m_cChat.m_nNewConnectionState = nState;
    }
    
    void OutputMsg( String sText, int nType )
    {
        if ( m_cChat != null )
            m_cChat.OutputMsg( sText, nType );
    }
    
    public String[] GetClients()
    {
        String[] aNicks = null;
        synchronized ( m_cClients )
        {
            if ( m_cClients.GetSize() > 0 )
            {
                aNicks = new String[ m_cClients.GetSize() ];
                for ( int i = 0; i<m_cClients.GetSize(); ++i )
                {
                    ChatClientInfo cInfo = m_cClients. GetClientInfoAt( i );
                    if ( cInfo == null )
                        continue;
                    aNicks[ i ] = cInfo.GetNick();                      
                }
            }
        }
        return aNicks;
    }
}

