/**
 * Manager statusu polaczenia
 * Dziala autonomicznie na kliencie.
 * Zbierane sa dane o poszczegolnych parametrach polaczenia.
 * Na ich podstawie wyswietlana jest ikonka ostrzegajaca
 * o ewentualnych problemach.
 * Gdy problemy sa powazne klient sam podejmuje decyzje
 * o odlaczeniu sie od serwera.
 */
public class ConnectionStatusManager extends GameObject
{
    /**
     * Status polaczenia a informacja na hudzie
     * ok      - brak ikonki
     * warning - ikonka miga
     * error   - ikonka wyswietla sie "na stale"
     */
    final static int _STATUS_OK             = 0,
                     _STATUS_WARNING        = 1, 
                     _STATUS_ERROR          = 2,
                     _STATUS_CRITICAL_ERROR = 3;
    
    final static int _NO_LOCK_PLAYER         = -1,
                     _LOCK_PLAYER_AT_WARNING =  _STATUS_WARNING,
                     _LOCK_PLAYER_AT_ERROR   =  _STATUS_ERROR;
    
    /**
     * Kolejne kryteria/komponenty oceny
     */
    final static int _LATENCY              = 0,
                     _PACKET_LOSS          = 1, 
                     _TIME_COMPENSATOR     = 2,
                     _COMPONENTS_CNT       = 3;
    
    
    
    final static int _COMPONENT_OK              = 0,
                     _COMPONENT_WARNING         = 1,
                     _COMPONENT_ERROR           = 2,
                     _COMPONENT_CRITICAL_ERROR  = 3;

    /**
     * Domyslny czas przez ktory musi trwac status
     * _STATUS_CRITICAL_ERROR (dla danego komponentu)
     * zeby gracz zostal odlaczony
     */
    final static int _DISCONNECT_DELAY_TIME     = 10;
    
    /**
     * Poziomy wartosci dla poszczegolnych kryteriow
     * Niektore z nich moga byc niemierzalne, to bedzie
     * je trzeba obsluzyc osobnymi metodami
     */
    final static ConnectionStatusInfo[] sm_aComponenstDef = 
                                                             new ConnectionStatusInfo [] 
    {
        //_LATENCY                                                             
        new ConnectionStatusInfo( 
                                  _LOCK_PLAYER_AT_ERROR,  //LOCK PLAYER
                                  400,  //WARNING THRESHOLD
                                  700,  //ERROR THRESHOLD
                                  1000,  //CRITICAL_ERROR THRESHOLD                                                                 
                                  10.0f //DISCONNECT_DELAY_TIME
                                ),
        //_PACKET_LOSS                                                             
        new ConnectionStatusInfo( 
                                  _LOCK_PLAYER_AT_ERROR,  //LOCK PLAYER
                                  10,  //WARNING THRESHOLD
                                  20,  //ERROR THRESHOLD
                                  40,  //CRITICAL_ERROR THRESHOLD
                                  10.0f //DISCONNECT_DELAY_TIME
                                ),                                                               
        //_TIME_COMPENSATOR                                                             
        new ConnectionStatusInfo(
                                  _LOCK_PLAYER_AT_ERROR //LOCK PLAYER
                                 )
    };
    
    /**
     * Aktualne poziomy bledow dla poszczegolnych kryteriow
     * Z tej tablicy bedzie wybierana najwyzsza wartosc
     * i wg niej bedzie wyswietlana ikonka na hudzie
     */
    final static byte[] m_aComponents = new byte[ _COMPONENTS_CNT ];
    
    
    
    /**
     * Obiekt game info, do ktorego bedziemy przesylali
     * dane o statusie polaczenia
     */
    NetGameInfoBase m_cGameInfo = null;
    
    
    /**
     * Aktualny status polaczenia,
     * jest to status calosciowy (wynikowy z poszczegolnych
     * komponentow, konkretnie max z wszystkich komponentow)
     */
    byte m_byCurrentStatus = _STATUS_OK;
    
    
    /**
     * Flaga mowiaca czy system juz wystartowac
     */
    private boolean m_bRunning = false;
    
       
    ConnectionStatusManager()
    {
        for ( int i = 0; i < m_aComponents.length; ++i )
            m_aComponents[ i ] = _STATUS_OK;
        m_byCurrentStatus = _STATUS_OK;
    }

    int METHOD_LockPlayer = 0;
    int METHOD_Disconnect = 0;
    
    void OnCreate()
    {
        METHOD_LockPlayer = RegisterCallMethod("LockPlayer");
        METHOD_Disconnect = RegisterCallMethod("Disconnect");
    }
    
    void OnDestroy()
    {
        m_cGameInfo = null;
    }
    
    void SetGameInfo(NetGameInfoBase cGameInfo)
    {
        m_cGameInfo = cGameInfo;
    }

    void Start()
    {
        m_bRunning = true;
    }
    
    void Stop()
    {
        for ( int nComponent = 0; nComponent < m_aComponents.length ; ++nComponent )
            SetStatus( nComponent, (byte)_STATUS_OK );
        CheckIfMainStatusChanged();
        m_bRunning = false;
    }
    
   
    /**
     * Interfejs do managera
     */
    
    /**
     * Wolane po aktualizacji pinga (latency) z serwera
     */
    void UpdateLatency( int nLatency )
    {
        Update( _LATENCY, nLatency );
    }
    
    /** Called when packet loss changes
     */
    void UpdatePacketLoss( int nPacketLoss )
    {
        Update( _PACKET_LOSS, nPacketLoss );
    }
    
    /**
     * Wolane po zmianach stanu compensatora czasu
     */
    void TimeCompensatorOk()
    {
        Update( _TIME_COMPENSATOR, _COMPONENT_OK );
    }
    void TimeCompensatorServerNotResponding()
    {
        Update( _TIME_COMPENSATOR, _COMPONENT_WARNING );
        
    }
    void TimeCompensatorConnectionToServerLost()
    {
        Update( _TIME_COMPENSATOR, _COMPONENT_ERROR );
    }
    
    
    
    /**
     * Prywatne metody przelaczajace status polaczenia
     * oraz obslugujace:
     * - huda
     * - blokowanie gracza
     * - odlaczanie od serwera 
     */
    
    
    
    /**
     * Update jednego z kryteriow (nComponent) na podsatawie 
     * wartosci podanej jako parametr.
     * Metoda ustala czy zmienil sie status dla tego kryterium, 
     * jesli tak, to jest on zapamietyuwany oraz wolana jest 
     * metoda CheckIfMainStatusChanged() (patrz nizej)
     */    
    private void Update( int nComponent, int nValue )
    {
        //przeliczmy nowy status tego komponentu
        byte byStatus = CalcStatus( nComponent, nValue );

        //zmienil sie od poprzedniego razu?
        if ( byStatus != GetStatus( nComponent ) )
        {
            //tak
            //NetLog.Test(" ChangeConnectionStatus "+nComponent+
            //            " ("+GetStatus( nComponent )+" -> "+byStatus+"\n");
            //zapamietajmy ten nowy status komponentu
            SetStatus( nComponent, byStatus );
            
            //i sprawdzmy czy to wplynelo na status calosciowy
            //jesli tak to uaktualnimy huda itp...
            CheckIfMainStatusChanged();
        }
    }

    /* 
     * Metoda przelicza status "calosciowy" m_byCurrentStatus 
     * i jesli on sie zmienil to wolane sa 
     * metody aktualizujace:
     * - huda
     * - blokowanie gracza
     * - odlaczanie od serwera
     */
    private void CheckIfMainStatusChanged()
    {
        if ( UpdateCurrentStatus() )
        {
            //tak, zmienil sie, zatem:
                
            //aktualizujemy stan ikonki na hudzie
            UpdateHud();
            //zmieniamy stan blokowania gracza (jesli taka potrzeba)
            if ( m_cGameInfo.IsLocalPlayer() )
                UpdateLockPlayer();
            //ewentualnie zaplanujmy/odwolajmy odlaczanie od gry 
            UpdateDisconnect();
        }
    }
    
    
    
    
    /** pobranie aktualnego statusu podanego komponentu*/
    private byte GetStatus(int nComponent) { return m_aComponents[ nComponent ]; }
    /** ustawienie statusu podanego komponentu*/
    private void SetStatus(int nComponent,byte byStatus) { m_aComponents[ nComponent ] = byStatus;}
    
    /** przeliczenie statusu dla podanego komponentu */
    private byte CalcStatus(int nComponent,int nValue)
    {
        int[] aThresholds = sm_aComponenstDef[ nComponent ].GetStatusThresholds();
        
        byte nStatus = _STATUS_OK;
        for ( ; nStatus < aThresholds.length; ++nStatus )
        {
            if ( nValue < aThresholds[ nStatus ] )
                return nStatus;
        }
        return nStatus;
    }
    
    
    
    /**
     * Metoda wolana po zmiania statusu jednego 
     * z komonnentow.
     * Przelicza ona czy zmienil sie status "calosciowy" m_byCurrentStatus
     * jesli tak, to zwraca true
     */
    private boolean UpdateCurrentStatus()
    {
        byte   byStatus             = _STATUS_OK;
        float  fDisconnectDelayTime = -1.0f;
        for ( int nComponent = 0 ; nComponent < m_aComponents.length ; ++nComponent )
        {
            if ( byStatus < m_aComponents[ nComponent ])
            {
                 byStatus = m_aComponents[ nComponent ];
                 
                 /*if ( byStatus < _STATUS_CRITICAL_ERROR )
                     continue;
                 
                 if ( (fDisconnectDelayTime < 0.0f) ||
                      (fDisconnectDelayTime < sm_aComponenstDef.GetDisconnectDelayTime()) )
                 {
                     fDisconnectDelayTime = sm_aComponenstDef.GetDisconnectDelayTime();
                 }*/
            }
        }
        
        if ( m_byCurrentStatus != byStatus )
        {
            m_byCurrentStatus = byStatus;
            return true;
        }
        
        return false;
    }
        
    /**
     * Aktualizacja ikonki na hudzie na podstawie aktualnego
     * statusu "calosciowego" m_byCurrentStatus
     */
    private void UpdateHud()
    {
        if ( m_cGameInfo.GetHUDNet() != null )
             m_cGameInfo.GetHUDNet().SetConnectionStatus( m_byCurrentStatus );
    }
    
    
    
    boolean m_bPlayerLocked = false;
    
    /**
     * Aktualizacja stanu blokowania gracza na podstawie 
     * danych w tablicy z bledami, wolane po zmianie poziomu 
     * bledu jednego z kryteriow/komponentow
     */
    private void UpdateLockPlayer()
    {
        boolean bPlayerLocked = false;
        for ( int nComponent = 0 ; nComponent < m_aComponents.length ; ++nComponent )
        {
            if ( (sm_aComponenstDef[ nComponent ].GetMinLockPlayerStatus() != _NO_LOCK_PLAYER)&&
                 (sm_aComponenstDef[ nComponent ].GetMinLockPlayerStatus() <=
                  m_aComponents[ nComponent ])
               )
            {
                bPlayerLocked = true;
                break;
            }
        }
        //czy zmienil sie stan blokowania gracza?
        if ( m_bPlayerLocked != bPlayerLocked )
        {
            //zaktualizujmy go
            m_bPlayerLocked = bPlayerLocked;
            //i zadzialajmy na gracza
            if ( m_bPlayerLocked )
                LockPlayer();
            else
                UnLockPlayer();
        }
    }
    
    /** Metoda wykonawcza blokujaca gracza */
    void LockPlayer()
    {
        DisableCallMethod( METHOD_LockPlayer );
        if ( m_cGameInfo.GetMyPlayerNet() != null )
            m_cGameInfo.GetMyPlayerNet().SetLockedByConnectionStatusManager( true );
        else            
            CallMethodIn( METHOD_LockPlayer, 0.5f );
    }
 
    /** Metoda wykonawcza odblokowujaca gracza */
    void UnLockPlayer()
    {
        DisableCallMethod( METHOD_LockPlayer );
        if ( m_cGameInfo.GetMyPlayerNet() != null )
             m_cGameInfo.GetMyPlayerNet().SetLockedByConnectionStatusManager( false );
    }
    
    
    /**
     * Aktualizacja stanu odlaczania klienta na podstawie
     * aktualnego statusu polaczenia
     */
    private void UpdateDisconnect()
    {
        if ( m_byCurrentStatus >= _STATUS_CRITICAL_ERROR )
            ScheduleDisconnect();
        else
            CancelDisconnect();
    }
    
    /** Metoda wykonawcza planujaca odlaczenie od gry/servera */
    void ScheduleDisconnect()
    {
        //dajemy troszke czasu opoznienia
        CallMethodIn( METHOD_Disconnect, _DISCONNECT_DELAY_TIME );
    }
    
    /** Metoda wykonawcza odlaczajaca od gry/servera */
    void Disconnect()
    {
        DisableCallMethod( METHOD_Disconnect );
        if ( m_cGameInfo != null )
            m_cGameInfo.EventDisconnectByConnectionStatusManager();
        else            
            CallMethodIn( METHOD_Disconnect, 0.1f );
    }
    
    /** Metoda wykonawcza odwolujaca odlaczenia od gry/servera */
    void CancelDisconnect()
    {
        DisableCallMethod( METHOD_Disconnect );
    }

}

class ConnectionStatusInfo
{
    private int    m_nLockPlayerStatus    = -1;
    private int [] m_aStatusThresholds    = null;
    private float  m_fDisconnectDelayTime = 0;//TODO: not used yet
    
    ConnectionStatusInfo()
    {
        this ( ConnectionStatusManager._NO_LOCK_PLAYER  );
    }
    ConnectionStatusInfo(int nLockPlayerStatus)
    {
        this( nLockPlayerStatus,
              ConnectionStatusManager._COMPONENT_WARNING,
              ConnectionStatusManager._COMPONENT_ERROR,
              ConnectionStatusManager._COMPONENT_CRITICAL_ERROR,
              ConnectionStatusManager._DISCONNECT_DELAY_TIME );
    }
    
    ConnectionStatusInfo(int   nLockPlayerStatus,
                         int   nWarningThreshold,
                         int   nErrorThreshold,
                         int   nCriticalErrorThreshold)
    {
        this( nLockPlayerStatus,
              nWarningThreshold,
              nErrorThreshold,
              nCriticalErrorThreshold,
              ConnectionStatusManager._DISCONNECT_DELAY_TIME );
    }
    
    ConnectionStatusInfo(int   nLockPlayerStatus,
                         int   nWarningThreshold,
                         int   nErrorThreshold,
                         int   nCriticalErrorThreshold,
                         float fDisconnectDelayTime)
    {
        m_nLockPlayerStatus = nLockPlayerStatus;
        
        m_aStatusThresholds = new int [3];
        m_aStatusThresholds[ 0 ] = nWarningThreshold;
        m_aStatusThresholds[ 1 ] = nErrorThreshold;
        m_aStatusThresholds[ 2 ] = nCriticalErrorThreshold;
        
        m_fDisconnectDelayTime = fDisconnectDelayTime;
    }
    
    int [] GetStatusThresholds(){ return m_aStatusThresholds; }
    int    GetMinLockPlayerStatus(){ return m_nLockPlayerStatus; }
    float  GetDisconnectDelayTime(){ return m_fDisconnectDelayTime; }
    
}