/** Klasa sprawdzajaca czy serwer odpowiada i nadzorujaca kompensatorem czasu
 */
public class NetTimeCompensator extends GameObject
{
    /** Co ile czasu sprawdzamy czy server odpowiada
     */
    //jesli poprzednio odpowiadal
    final static float _CHECK_SERVER_RESPONDING_TIME_IF_YES  = 5.0f;
    //jesli poprzednio nie odpowiadal 
    final static float _CHECK_SERVER_RESPONDING_TIME_IF_NO   = 0.5f;
    
    /**
     * Przez jaki czas serwer musi nie odpowiadac, zeby uznac
     * ze stracilismy polaczenie
     */
    final static float _MIN_TIME_TO_SUSPECT_CONNECTION_LOST = 45.0f;
        
    NetGameInfoBase m_cGameInfo = null;
    
    void SetGameInfo( NetGameInfoBase cGameInfo )
    {
        m_cGameInfo = cGameInfo;
    }
    
    int METHOD_DelayedStart             = 0;
    int METHOD_SendServerPing           = 0;
    int METHOD_CheckServerResponding    = 0;
    int METHOD_StartCompensingTime      = 0;
    
    public void OnCreate()
    {
        METHOD_DelayedStart             = RegisterCallMethod("DelayedStart");
        METHOD_SendServerPing           = RegisterCallMethod("SendServerPing");
        METHOD_CheckServerResponding    = RegisterCallMethod("CheckServerResponding");
        METHOD_StartCompensingTime      = RegisterCallMethod("StartCompensingTime");
    }
    
    public void OnDestroy()
    {
        m_cGameInfo = null;
    }
    
    /** Start systemu
     */
    void Start()
    {
        CallMethodEvery( METHOD_SendServerPing,        
                         _CHECK_SERVER_RESPONDING_TIME_IF_YES * 0.21f );        
        
        CallMethodIn   ( METHOD_DelayedStart, 
                         _CHECK_SERVER_RESPONDING_TIME_IF_YES * 2.0f );
    }
    
    void DelayedStart()
    {
        StartCompensingTime();
        m_fLastServerRespondedTime = GetTime();
        CheckServerResponding();
    }
     
    void Stop()
    {
        StopCompensingTime();
        DisableCallMethod( METHOD_DelayedStart );
        DisableCallMethod( METHOD_CheckServerResponding );        
        DisableCallMethod( METHOD_SendServerPing );        
    }
    
    boolean m_bLastServerResponded  = true;
    boolean m_bCompensingTime       = false;
    
    /** Czas w ktorym mielismy ostatnio znak zycia od serwera
     */
    float m_fLastServerRespondedTime = 0.0f;
    
    /** Rozpoczecie kompensacji czasu
     */
    void StartCompensingTime()
    {
        if ( ! m_bCompensingTime )
        {
            m_bCompensingTime = true;
            super.StartCompensingTime();
        }
    }
    
    /** Wstrzymanie kompensacji czasu
     */
    void StopCompensingTime()
    {
        DisableCallMethod( METHOD_StartCompensingTime );
        if ( m_bCompensingTime )
        {
            m_bCompensingTime = false;
            super.StopCompensingTime();
        }
    }
    
    /** Sprawdzenie czy serwer odpowiada
     */
    void CheckServerResponding()
    {
        //wstepnie ustalmy czas kolejnego sprawdzenia serwera
        float fCheckServerRespondingTime = _CHECK_SERVER_RESPONDING_TIME_IF_YES;
        
        //odpowiada?
        if ( IsServerResponding() )
        {
            //tak, 
            //zapamietajmy sobie ten czas
            m_fLastServerRespondedTime = GetTime();
                
            //a co bylo ostattnio?
            if ( ! m_bLastServerResponded )
            {
                //ostatnio nie odpowiadal...
                
                //powiadamiamy o tym NetGameInfo, zeby
                //m.in. schowac komunikat ostrzegawczy
                if ( m_cGameInfo != null )
                    m_cGameInfo.EventServerResponding();
                                
                //ustalmy czas kolejnego sprawdzenia serwera
                //teraz krotszy niz zwykle bo trzeba go miec na uwagdze
                fCheckServerRespondingTime = 0.5f * _CHECK_SERVER_RESPONDING_TIME_IF_YES;
                
                //za jakis czas zaczniemy kompensacje czasu, oczywiscie jak
                //do tego czasu serwer nie "zamilknie"
                CallMethodIn( METHOD_StartCompensingTime, 
                              2.0f * _CHECK_SERVER_RESPONDING_TIME_IF_YES );
                
                //zapamietajmy ze serwer aktualnei odpowiada
                m_bLastServerResponded = true;
            }
        }
        else
        {
            //nie, 
            //a co bylo ostatnio?
            if ( m_bLastServerResponded )
            {
                //ostatnio odpowiadal...
                
                //wyswietlmy komunikat ostrzegawczy
                if ( m_cGameInfo != null )
                    m_cGameInfo.EventServerNotResponding();
                
                //przerwijmy kompensacje czasu (takze ta zaplanowana w CallMethodIn)
                StopCompensingTime();
                
                //zapamietajmy ze serwer aktualnie nie odpowiada
                m_bLastServerResponded = false;
            }
            else
            {
                //ostatnio tez nie odpowiadal, 
                //ciekawe jak dlugo juz milczy
                float fDelta = GetTime() - m_fLastServerRespondedTime;
                
                if ( fDelta >= _MIN_TIME_TO_SUSPECT_CONNECTION_LOST )
                {
                    if ( m_cGameInfo != null )
                         m_cGameInfo.EventConnectionToServerLost();
                }
            }
            
            //ustalmy czas kolejnego sprawdzenia serwera
            fCheckServerRespondingTime = _CHECK_SERVER_RESPONDING_TIME_IF_NO;
        }
        
        //zaplanujmy kolejne sprawdzenie
        CallMethodIn( METHOD_CheckServerResponding, 
                      fCheckServerRespondingTime );
    }
    
    /** Poprosimy NetGameInfo zeby "pingnelo" serwer
     */
    void SendServerPing()
    {
        if ( m_cGameInfo != null )
            m_cGameInfo.MsgSendServerPing();
    }
    
    String GetInfo()
    {
        boolean bIsCompensing = IsCompensingTime();
        String sRet = ""+bIsCompensing;
        if ( bIsCompensing )
            sRet += " Time: "+GetTimeToCompense()+
                    " LastTime(): "+GetLastTimeToCompense();
        return sRet;
    }
}
